#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <iostream>
#include <cstdlib>

#define PRECISION 40
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define ABS(a, b) (((a) > (b)) ? ((a)-(b)) : ((b)-(a)))


using namespace std;


/*
branchFactor() receives in input a branching vector
It returns the corresponding branching factor "bf"
This is done via doubling + binary search. The desired value satisfies 
1-sum_{j=0}^{n-1}bf^(-V[i])=0
The binary search is anyway interrupted when the value is accurate enough 
*/
double branchFactor(int n, double* V){

  double left=0.0;
  double right=1.0;
  double f;
  //we compute an upper bound right on the branching factor via doubling
  do{
    right = right*2;   
    f=1.0;
    for(int j=0; j<n; j++) {
      f -= pow(right, -V[j]);
    }          
  }while(f<=0);       
  //we compute the branching factor via binary search
  double bf;
  for(int i=0; i<PRECISION; i++) {
    bf = (left+right)/2.0;
    f=1.0;
    for(int j=0; j<n; j++) {
      f -= pow(bf, -V[j]);
    }
    if(f==0) return bf;
    if(f>0) right = bf;
    else left = bf;
  }
  return right;//this way we return an upper bound            
}



/*
lambda is the value of the branching factor for given (feasible) weights. 
"stampa" is used to print or not some details
*/
double computeLambda(double a, int stampa) {
  
  
  double V[10];//we assume that we branch on at most 10 subproblem, which is the case here
  double lambda, lambda_max= -1;

  //We don't need to consider the base case and folding

  //Branch at degree >=5
  //n3 and n4 are the neighbors of degree 3 and 4
  //n5 are the neighbors of degree at least 5
  for(int n3=0; n3<=5; n3++){
    for(int n4=0; n4<=5-n3; n4++){      
      int n5 = 5-n3-n4;
      V[0]=1+n3*a+n4*(1-a)+n5*0;
      V[1]=1+n3*a+n4*1+n5*1;
      lambda = branchFactor(2, V);
      if(stampa) printf("branch at 5: 1+%d*a+%d*(1-a)+%d*0/1+%d*a+%d*1+%d*1=%.8lf\n", n3, n4, n5, n3, n4, n5, lambda);
      lambda_max = MAX(lambda_max, lambda);    
    }
  }

  //Branch at degree 4
  for(int n3=0; n3<=4; n3++){
    int n4 = 4-n3;
    V[0]=1+n3*a+n4*(1-a);
    V[1]=1+n3*a+n4*1;
    lambda = branchFactor(2, V);
    if(stampa) printf("branch at 4: 1+%d*a+%d*(1-a)/1+%d*a+%d*1=%.8lf\n", n3, n4, n3, n4, lambda);
    lambda_max = MAX(lambda_max, lambda);    
  }


  //Branch at degree 3
  V[0]=4*a;
  V[1]=4*a;
  lambda = branchFactor(2, V);
  if(stampa) printf("branch at 3: 4a/4a=%.8lf\n", lambda);
  lambda_max = MAX(lambda_max, lambda);
  
   
  if(stampa) printf("\nlambda_max(a=%.8lf)=%.8lf\n", a, lambda_max);
  
  return lambda_max;
}



//Here we put the constraints on the weights: in our case a\in [0.5,1.0]
bool feasible(double a){
  if(a>=0.5 && a<=1.0) return true;
  else return false;                        
}

int main() {

  double lambdamin = 1000;
  double amin = 1000;  
  
  //we search for the best a in a grid with offset 1/grid         
  int grid=100;
  for(int i=1; i<=grid; i++){
    double a=(double)i/grid;
    if(feasible(a)){
      double lambda = computeLambda(a,0);
      printf("lambda(a=%lf)=%.8lf\n", a, lambda);              
      if(lambda < lambdamin){
        lambdamin = lambda;
        amin = a;      
      }
    }
  }       
  printf("\n\nBEST BOUND\n");  
  computeLambda(amin,1);
//  printf("\nlambdamin(amin=%lf)=%.8lf\n", amin, lambdamin);              


  system("PAUSE");
  return 0;

}
