#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <iostream>
#include <cstdlib>

#define PRECISION 40
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define ABS(a, b) (((a) > (b)) ? ((a)-(b)) : ((b)-(a)))


using namespace std;


/*
branchFactor() receives in input a branching vector
It returns the corresponding branching factor "bf"
This is done via doubling + binary search. The desired value satisfies 
1-sum_{j=0}^{n-1}bf^(-V[i])=0
The binary search is anyway interrupted when the value is accurate enough 
*/
double branchFactor(int n, double* V){

  double left=0.0;
  double right=1.0;
  double f;
  //we compute an upper bound right on the branching factor via doubling
  do{
    right = right*2;   
    f=1.0;
    for(int j=0; j<n; j++) {
      f -= pow(right, -V[j]);
    }          
  }while(f<=0);       
  //we compute the branching factor via binary search
  double bf;
  for(int i=0; i<PRECISION; i++) {
    bf = (left+right)/2.0;
    f=1.0;
    for(int j=0; j<n; j++) {
      f -= pow(bf, -V[j]);
    }
    if(f==0) return bf;
    if(f>0) right = bf;
    else left = bf;
  }
  return right;//this way we return an upper bound            
}

double uniform(double left, double right) {
  return left + (right-left)*((double)rand())/RAND_MAX;
}



/*
lambda is the value of the branching factor for given (feasible) weights. 
"stampa" is used to print or not some details
*/
double computeLambda(double a3, double a4, int stampa) {
  
  
  double V[10];//we assume that we branch on at most 10 subproblem, which is the case here
  double lambda, lambda_max= -1;

  //We don't need to consider the base case and folding

  //Branch at degree >=6
  //Here n6 are the neighbors of degree at least 6
  for(int n3=0; n3<=6; n3++){
    for(int n4=0; n4<=6-n3; n4++){      
      for(int n5=0; n5<=6-n3-n4; n5++){      
        int n6 = 6-n3-n4-n5;
        V[0]=1+n3*a3+n4*(a4-a3)+n5*(1-a4)+n6*0;
        V[1]=1+n3*a3+n4*a4+n5*1+n6*1;
        lambda = branchFactor(2, V);
        if(stampa) printf("branch at 6: 1+%d*a3+%d*(a4-a3)+%d*(1-a4)+%d*0/1+%d*a3+%d*a4+%d*1+%d*1=%.8lf\n", 
                           n3, n4, n5, n6, n3, n4, n5, n6, lambda);
        lambda_max = MAX(lambda_max, lambda);    
      }
    }
  }

  //Branch at degree 5
  for(int n3=0; n3<=5; n3++){
    for(int n4=0; n4<=5-n3; n4++){      
      int n5 = 5-n3-n4;
      V[0]=1+n3*a3+n4*(a4-a3)+n5*(1-a4);
      V[1]=1+n3*a3+n4*a4+n5*1;
      lambda = branchFactor(2, V);
      if(stampa) printf("branch at 5: 1+%d*a3+%d*(a4-a3)+%d*(1-a4)/1+%d*a3+%d*a4+%d*1=%.8lf\n", 
                           n3, n4, n5, n3, n4, n5, lambda);
      lambda_max = MAX(lambda_max, lambda);    
    }
  }

  //Branch at degree 4
  for(int n3=0; n3<=4; n3++){      
    int n4 = 4-n3;
    V[0]=a4+n3*a3+n4*(a4-a3);
    V[1]=a4+n3*a3+n4*a4;
    lambda = branchFactor(2, V);
    if(stampa) printf("branch at 4: a4+%d*a3+%d*(a4-a3)/a4+%d*a3+%d*a4=%.8lf\n", 
                           n3, n4, n3, n4, lambda);
    lambda_max = MAX(lambda_max, lambda);    
  }

  //Branch at degree 3
  int n3 = 3;
  V[0]=a3+n3*a3;
  V[1]=a3+n3*a3;
  lambda = branchFactor(2, V);
  if(stampa) printf("branch at 3: a3+%d*a3/a3+%d*a3=%.8lf\n", 
                           n3, n3,lambda);
  lambda_max = MAX(lambda_max, lambda);    

   
  if(stampa) printf("\nlambda_max(a3=%.8lf,a4=%.8lf)=%.8lf\n", a3, a4, lambda_max);
  
  return lambda_max;
}



//Here we put the constraints on the weights
bool feasible(double a3, double a4){
  if(a4<=1 && a3<=a4 && a3>0 && 2*a3>=a4 && 2*a4>=1 && a4>=1-a3) return true;
  else return false;                        
}

int main() {

  double lambda_min;
  double a3min;  
  double a4min;  

  //compute an initial random solution
  do{
    a3min = uniform(0.0,1.0);
    a4min = uniform(0.0,1.0);          
  }while(!feasible(a3min,a4min));           
  lambda_min = computeLambda(a3min,a4min,0);          
  printf("lambda(a3=%lf,a4=%lf)=%.8lf\n", a3min, a4min, lambda_min);              


  double deltamin = 0.0001; //final step
  double delta = 0.01; //initial step
  int counter = 0; //measures for how long we don't make any progress
  do{
    double a3rand = a3min + delta*uniform(-1.0,1.0);
    double a4rand = a4min + delta*uniform(-1.0,1.0);   
    counter++; //a new proposal is generated
    if(feasible(a3rand,a4rand)) {
      double lambda_rand = computeLambda(a3rand,a4rand,0);
      if(lambda_rand < lambda_min) {
        counter = 0;
        lambda_min = lambda_rand;
        a3min = a3rand;
        a4min = a4rand;
        printf("lambda(a3=%lf,a4=%lf)=%.8lf\n", a3min, a4min, lambda_min);              
      }
    }
    if(counter >= 1000){
      counter = 0;
      delta = delta/2.0;
      printf("\ndelta=%.10f\n\n", delta);
    }
  }while(delta >= deltamin);
  
  printf("\n\nBEST BOUND\n");  
  computeLambda(a3min,a4min,1);

  system("PAUSE");
  return 0;

}
